#include <string.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include "undo.h"
#include "indent.h"

static gboolean auto_indent = FALSE;

void indent_change_state(gboolean state)
{
	auto_indent = state;
}

static gchar *compute_indentation(GtkTextBuffer *buffer, gint line) // from gedit
{
	GtkTextIter start_iter, end_iter;
	gunichar c;
	
	gtk_text_buffer_get_iter_at_line(buffer, &start_iter, line);
	end_iter = start_iter;
	c = gtk_text_iter_get_char(&end_iter);
	while (g_unichar_isspace(c) && c != '\n') {
		if (!gtk_text_iter_forward_char(&end_iter))
			break;
		c = gtk_text_iter_get_char(&end_iter);
	}
	if (gtk_text_iter_equal(&start_iter, &end_iter))
		return NULL;
	
	return gtk_text_iter_get_text(&start_iter, &end_iter);
}

static void indent_real(GtkTextBuffer *buffer)
{
	GtkTextIter iter;
	gchar *ind, *str;
	
	gtk_text_buffer_get_iter_at_mark(buffer, &iter, gtk_text_buffer_get_insert(buffer));
	ind = compute_indentation(buffer, gtk_text_iter_get_line(&iter));
	str = g_strconcat("\n", ind, NULL);
	gtk_text_buffer_insert(buffer, &iter, str, -1);
	g_free(str);
	g_free(ind);
}

static void multi_line_indent(GtkTextBuffer *buffer)
{
	GtkTextIter start_iter, end_iter, iter;
	gint start_line, end_line, i;
	gboolean pos;
	
	gtk_text_buffer_get_selection_bounds(buffer, &start_iter, &end_iter);
	start_line = gtk_text_iter_get_line(&start_iter);
	end_line = gtk_text_iter_get_line(&end_iter);
	gtk_text_buffer_get_iter_at_mark(buffer, &iter, gtk_text_buffer_get_insert(buffer));
	pos = gtk_text_iter_equal(&iter, &start_iter);
	for (i = start_line; i < end_line; i++) {
		gtk_text_buffer_get_iter_at_line(buffer, &iter, i);
		gtk_text_buffer_place_cursor(buffer, &iter);
		gtk_text_buffer_insert(buffer, &iter, "\t", 1);
		undo_set_sequency(TRUE);
	}
	undo_set_sequency(FALSE);
	
	gtk_text_buffer_get_iter_at_line(buffer, &start_iter, start_line);
	gtk_text_buffer_get_iter_at_line(buffer, &end_iter, end_line);
	if (pos) {
		gtk_text_buffer_place_cursor(buffer, &end_iter);
		gtk_text_buffer_move_mark_by_name(buffer, "insert", &start_iter);
	} else {
		gtk_text_buffer_place_cursor(buffer, &start_iter);
		gtk_text_buffer_move_mark_by_name(buffer, "insert", &end_iter);
	}
}

static gint compute_indent_offset_length(const gchar *ind)
{
	guint8 c = *ind++;
	gint len = 1;
	
	while ((len < 8) && (c = *ind++) == 0x20) {
		len++;
	}
	
	return len;
}

static void multi_line_unindent(GtkTextBuffer *buffer)
{
	GtkTextIter start_iter, end_iter, iter;
	gint start_line, end_line, i, len;
	gboolean pos;
	gchar *ind;
	
	gtk_text_buffer_get_selection_bounds(buffer, &start_iter, &end_iter);
	start_line = gtk_text_iter_get_line(&start_iter);
	end_line = gtk_text_iter_get_line(&end_iter);
	gtk_text_buffer_get_iter_at_mark(buffer, &iter, gtk_text_buffer_get_insert(buffer));
	pos = gtk_text_iter_equal(&iter, &start_iter);
//	for (i = start_line; i < end_line; i++) {
	i = start_line;
	do {
		ind = compute_indentation(buffer, i);
		if (ind && strlen(ind)) {
			len = compute_indent_offset_length(ind);
			gtk_text_buffer_get_iter_at_line(buffer, &start_iter, i);
			gtk_text_buffer_place_cursor(buffer, &start_iter);
			end_iter = start_iter;
			gtk_text_iter_forward_chars(&end_iter, len);
			gtk_text_buffer_move_mark_by_name(buffer, "insert", &end_iter);
			gtk_text_buffer_delete(buffer, &start_iter, &end_iter);
			undo_set_sequency(TRUE);
			g_free(ind);
		}
		i++;
	} while (i < end_line);
//	}
	undo_set_sequency(FALSE);
	
	gtk_text_buffer_get_iter_at_line(buffer, &start_iter, start_line);
	gtk_text_buffer_get_iter_at_line(buffer, &end_iter, end_line);
	if (pos) {
		gtk_text_buffer_place_cursor(buffer, &end_iter);
		gtk_text_buffer_move_mark_by_name(buffer, "insert", &start_iter);
	} else {
		gtk_text_buffer_place_cursor(buffer, &start_iter);
		gtk_text_buffer_move_mark_by_name(buffer, "insert", &end_iter);
	}
}

static gboolean check_preedit(GtkWidget *text_view)
{
	gchar *str;
	
	gtk_im_context_get_preedit_string(
		GTK_TEXT_VIEW(text_view)->im_context, &str, NULL, NULL);
	if (strlen(str)) {
		g_free(str);
		return TRUE;
	}
	g_free(str);
	return FALSE;
}

static gboolean check_selection_bound(GtkTextBuffer *buffer)
{
	GtkTextIter start_iter, end_iter;
	gchar *str;
	
	if (gtk_text_buffer_get_selection_bounds(buffer, &start_iter, &end_iter)) {
		str = gtk_text_iter_get_text(&start_iter, &end_iter);
		if (g_strrstr(str, "\n")) {
			g_free(str);
			return TRUE;
		}
		g_free(str);
	}
	return FALSE;
}

static gboolean cb_key_press_event(GtkWidget *text_view, GdkEventKey *event)
{
	switch (event->keyval) {
	case GDK_Return:
		if (check_preedit(text_view))
			return FALSE;
		if ((auto_indent && !(event->state &= GDK_SHIFT_MASK)) ||
			(!auto_indent && (event->state &= GDK_SHIFT_MASK))) {
			indent_real(gtk_text_view_get_buffer(GTK_TEXT_VIEW(text_view)));
			return TRUE;
		}
		break;
	case GDK_Tab:
	case GDK_ISO_Left_Tab:
		{
			GtkTextBuffer *buffer =
				gtk_text_view_get_buffer(GTK_TEXT_VIEW(text_view));
			
			if (event->state &= GDK_SHIFT_MASK)
				multi_line_unindent(buffer);
			else if (!check_selection_bound(buffer))
				break;
			else
				multi_line_indent(buffer);
			return TRUE;
		}
	}
	return FALSE;
}

void indent_init(GtkWidget *text_view)
{
	g_signal_connect(G_OBJECT(text_view), "key-press-event",
		G_CALLBACK(cb_key_press_event), NULL);
}
