#include <gtk/gtk.h>
#include "undo.h"

#define DV(x)

typedef struct {
	gint command;
	gint start;
	gint end;
	gboolean seq; // sequency flag
	gchar *str;
} UndoInfo;

enum {
	DELETE = 0,
	INSERT
};

static GList *undo_list = NULL;
static GList *redo_list = NULL;
static gint step_modif;
static GtkWidget *view;

static void undo_clear_undo_info(void)
{
	while (g_list_length(undo_list)) {
		g_free(((UndoInfo *)undo_list->data)->str);
		g_free(undo_list->data);
		undo_list = g_list_delete_link(undo_list, undo_list);
	}
DV(g_print("undo_cb: Undo list cleared by %d\n", g_list_length(undo_list)));
}

static void undo_clear_redo_info(void)
{
	while (g_list_length(redo_list)) {
		g_free(((UndoInfo *)redo_list->data)->str);
		g_free(redo_list->data);
		redo_list = g_list_delete_link(redo_list, redo_list);
	}
DV(g_print("undo_cb: Redo list cleared by %d\n", g_list_length(redo_list)));
}

static void undo_append_undo_info(GtkTextBuffer *buffer, gint command, gint start, gint end, gboolean seq)
{
	GtkTextIter start_iter, end_iter;
	UndoInfo *ui = g_malloc(sizeof(UndoInfo));
	
	gtk_text_buffer_get_iter_at_offset(buffer, &start_iter, start);
	gtk_text_buffer_get_iter_at_offset(buffer, &end_iter, end);
	ui->command = command;
	ui->start = start;
	ui->end = end;
	ui->seq = seq;
	ui->str = gtk_text_buffer_get_text(buffer, &start_iter, &end_iter, FALSE);
	undo_list = g_list_append(undo_list, ui);
DV(g_print("undo_cb: %d %s (%d-%d) %d\n", command, ui->str, start, end, seq));
	undo_clear_redo_info();
}

static void cb_delete_range(GtkTextBuffer *buffer)
{
	GtkTextIter start_iter, end_iter;
	gint start, end;
	
	gtk_text_buffer_get_selection_bounds(buffer, &start_iter, &end_iter);
	start = gtk_text_iter_get_offset(&start_iter);
	end = gtk_text_iter_get_offset(&end_iter);
	if (start == end)
		start--;
	undo_append_undo_info(buffer, DELETE, start, end, FALSE);
}

static void cb_insert_text(GtkTextBuffer *buffer)
{
	GtkTextIter start_iter, end_iter;
	gboolean seq = FALSE;
	static gint start, end;
	static gboolean after = FALSE;
	
	gtk_text_buffer_get_selection_bounds(buffer, &start_iter, &end_iter);
	if (!after) {
		start = gtk_text_iter_get_offset(&start_iter);
		after = TRUE;
	} else {
		end = gtk_text_iter_get_offset(&start_iter);
		after = FALSE;
		if (gtk_text_iter_compare(&start_iter, &end_iter) != 0)
			seq = TRUE; // for selection bound over write
		undo_append_undo_info(buffer, INSERT, start, end, seq);
	}
}

static void set_main_window_title_with_asterisk(gboolean flag)
{
	const gchar *old_title =
		gtk_window_get_title(GTK_WINDOW(gtk_widget_get_toplevel(view)));
	gchar *new_title;
	
	if (flag)
		new_title = g_strconcat("*", old_title, NULL);
	else {
		GString *gstr = g_string_new(old_title);
		gstr = g_string_erase(gstr, 0, 1);
		new_title = g_string_free(gstr, FALSE);
	}
	gtk_window_set_title(GTK_WINDOW(gtk_widget_get_toplevel(view)), new_title);
	g_free(new_title);
}
/*
static void set_main_window_title_toggle_asterisk(void)
{
	const gchar *old_title =
		gtk_window_get_title(GTK_WINDOW(gtk_widget_get_toplevel(view)));
	gchar *new_title;
	
	if (g_str_has_prefix(old_title, "*")) {
		GString *gstr = g_string_new(old_title);
		gstr = g_string_erase(gstr, 0, 1);
		new_title = g_string_free(gstr, FALSE);
	} else {
		new_title = g_strconcat("*", old_title, NULL);
	}
	gtk_window_set_title(GTK_WINDOW(gtk_widget_get_toplevel(view)), new_title);
	g_free(new_title);
}
*/
static void cb_modified_changed(void)
{
	set_main_window_title_with_asterisk(TRUE);
}

void undo_reset_step_modif(void)
{
	step_modif = g_list_length(undo_list);
DV(g_print("undo_reset_step_modif: Reseted step_modif by %d\n", step_modif));
}

static void undo_check_step_modif(GtkTextBuffer *buffer)
{
	if (g_list_length(undo_list) == step_modif) {
		g_signal_handlers_block_by_func(G_OBJECT(buffer), G_CALLBACK(cb_modified_changed), NULL);
		gtk_text_buffer_set_modified(buffer, FALSE);
		g_signal_handlers_unblock_by_func(G_OBJECT(buffer), G_CALLBACK(cb_modified_changed), NULL);
		set_main_window_title_with_asterisk(FALSE);
	}
}

static gint undo_connect_signal(GtkTextBuffer *buffer)
{
	g_signal_connect(G_OBJECT(buffer), "delete-range",
		G_CALLBACK(cb_delete_range), buffer);
	g_signal_connect(G_OBJECT(buffer), "insert-text",
		G_CALLBACK(cb_insert_text), buffer);
	g_signal_connect_after(G_OBJECT(buffer), "insert-text",
		G_CALLBACK(cb_insert_text), buffer);
	return 
	g_signal_connect(G_OBJECT(buffer), "modified-changed",
		G_CALLBACK(cb_modified_changed), NULL);
}

void undo_init(GtkWidget *textview, GtkTextBuffer *buffer)
{
	static guint flag = 0;
	
	if (undo_list)
		undo_clear_undo_info();
	if (redo_list)
		undo_clear_redo_info();
	undo_reset_step_modif();
DV(g_print("undo_init: list reseted\n"));
	
	if (!flag) {
		view = textview;
		flag = undo_connect_signal(buffer);
	}
}

gint undo_block_signal(GtkTextBuffer *buffer)
{
	return 
	g_signal_handlers_block_by_func(G_OBJECT(buffer), G_CALLBACK(cb_delete_range), buffer) +
//	g_signal_handlers_block_by_func(G_OBJECT(buffer), G_CALLBACK(cb_modified_changed), buffer) +
	g_signal_handlers_block_by_func(G_OBJECT(buffer), G_CALLBACK(cb_insert_text), buffer);
}

gint undo_unblock_signal(GtkTextBuffer *buffer)
{
	return 
	g_signal_handlers_unblock_by_func(G_OBJECT(buffer), G_CALLBACK(cb_delete_range), buffer) +
//	g_signal_handlers_unblock_by_func(G_OBJECT(buffer), G_CALLBACK(cb_modified_changed), buffer) +
	g_signal_handlers_unblock_by_func(G_OBJECT(buffer), G_CALLBACK(cb_insert_text), buffer);
}

gint undo_disconnect_signal(GtkTextBuffer *buffer) //must be not needed
{
	return 
	g_signal_handlers_disconnect_by_func(G_OBJECT(buffer), G_CALLBACK(cb_delete_range), buffer) +
//	g_signal_handlers_disconnect_by_func(G_OBJECT(buffer), G_CALLBACK(cb_modified_changed), buffer) +
	g_signal_handlers_disconnect_by_func(G_OBJECT(buffer), G_CALLBACK(cb_insert_text), buffer);
}

void undo_set_sequency(gboolean seq)
{
	UndoInfo *ui;
	
	if (g_list_length(undo_list)) {
		ui = g_list_last(undo_list)->data;
		ui->seq = seq;
	}
}

gboolean undo_undo(GtkTextBuffer *buffer)
{
	GtkTextIter start_iter, end_iter;
	UndoInfo *ui;
	
	if (g_list_length(undo_list)) {
		undo_block_signal(buffer);
		ui = g_list_last(undo_list)->data;
		gtk_text_buffer_get_iter_at_offset(buffer, &start_iter, ui->start);
		switch (ui->command) {
		case DELETE:
			gtk_text_buffer_insert(buffer, &start_iter, ui->str, -1);
			break;
		case INSERT:
			gtk_text_buffer_get_iter_at_offset(buffer, &end_iter, ui->end);
			gtk_text_buffer_delete(buffer, &start_iter, &end_iter);
			break;
		}
		redo_list = g_list_append(redo_list, ui);
		undo_list = g_list_delete_link(undo_list, g_list_last(undo_list));
DV(g_print("cb_edit_undo: undo list left %d\n", g_list_length(undo_list)));
		undo_unblock_signal(buffer);
		undo_check_step_modif(buffer);
		if (g_list_length(undo_list))
			if (((UndoInfo *)g_list_last(undo_list)->data)->seq)
				return TRUE;
		gtk_text_buffer_place_cursor(buffer, &start_iter);
		gtk_text_view_scroll_to_iter(GTK_TEXT_VIEW(view), &start_iter,
			0.1, FALSE, 0.5, 0.5);
	}
	return FALSE;
}

gboolean undo_redo(GtkTextBuffer *buffer)
{
	GtkTextIter start_iter, end_iter;
	UndoInfo *ri;
	
	if (g_list_length(redo_list)) {
		undo_block_signal(buffer);
		ri = g_list_last(redo_list)->data;
		gtk_text_buffer_get_iter_at_offset(buffer, &start_iter, ri->start);
		switch (ri->command) {
		case DELETE:
			gtk_text_buffer_get_iter_at_offset(buffer, &end_iter, ri->end);
			gtk_text_buffer_delete(buffer, &start_iter, &end_iter);
			break;
		case INSERT:
			gtk_text_buffer_insert(buffer, &start_iter, ri->str, -1);
			break;
		}
		undo_list = g_list_append(undo_list, ri);
		redo_list = g_list_delete_link(redo_list, g_list_last(redo_list));
DV(g_print("cb_edit_redo: redo list left %d\n", g_list_length(redo_list)));
		undo_unblock_signal(buffer);
		undo_check_step_modif(buffer);
		if (ri->seq)
			return TRUE;
		gtk_text_buffer_place_cursor(buffer, &start_iter);
		gtk_text_view_scroll_to_iter(GTK_TEXT_VIEW(view), &start_iter,
			0.1, FALSE, 0.5, 0.5);
	}
	return FALSE;
}
